from pathlib import Path
from typing import Sequence, Tuple, Union


# https://matplotlib.org/stable/users/explain/colors/colors.html
ColorArg = Union[str, Sequence[float], Tuple[str, float]]
PathArg = Union[Path, str]



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""
Reuires https://github.com/maenu/scholar.py::

    pip install https://github.com/maenu/scholar.py/archive/master.zip

!! Run from interpreter to retain data for debugging access

run find_publications.py with command

"""
from argparse import ArgumentParser
from collections import defaultdict
import json
from pathlib import Path
import pickle
import random
import re
import time
from typing import List

from eelbrain._utils import ask
from pybtex.database import BibliographyData, parse_bytes, parse_file
import requests


# Query
API_KEY_PATH = Path('~/.serpapi_key').expanduser()
API_KEY = API_KEY_PATH.read_text().strip()
DIR = Path(__file__).absolute().parent
RAW_SEARCH_CACHE = DIR / 'cache-search.json'
SEARCH_CACHE = DIR / 'cache.json'
SELECTION_PATH = DIR / 'selection.json'
BIBTEX_CACHE = DIR / 'bibtex.pickle'
# Parse
DST = DIR.parent / 'publications.bib'
OBSOLETE = """
Qm4WZWrznhEJ
OdvLIq-CqsUJ
VLeSaj8CLlcJ
ZnxohaAY4fEJ
""".split()
BIORXIV_OBSOLETE = """
10.1101/326785
2018/10/09/439158
10.1101/2020.08.17.254482
10.1101/421727
10.1101/2021.10.15.464457
""".split()
ACRONYMS = ['EEG', 'MEG', 'MRI']


def download_entries() -> List:
    entries = []
    # download entries
    total_results = None
    num = 20
    start = 0
    while True:
        r = requests.get(f"https://serpapi.com/search.json?engine=google_scholar&q=%22eelbrain%22&as_ylo=2015&hl=en&start={start}&num={num}&as_vis=1&api_key={API_KEY}")
        content = json.loads(r.content)
        if not total_results:
            total_results = content['search_information']['total_results']
        entries.extend(content['organic_results'])
        # next page
        start += num
        if start >= total_results:
            break
    return entries


def search(cached: bool):
    # Raw entry list from Google
    if cached:
        with RAW_SEARCH_CACHE.open('rt') as file:
            entries = json.load(file)
    else:
        entries = download_entries()
        with RAW_SEARCH_CACHE.open('wt') as file:
            json.dump(entries, file)
    # One entry per key
    entry_dict = {e['result_id']: e for e in entries}
    if len(entry_dict) != len(entries):
        counter = defaultdict(list)
        for e in entries:
            counter[e['result_id']].append(e)
        duplicates = {k: k_entries for k, k_entries in counter.items() if len(k_entries) > 1}
        print("Warning: duplicate entries for keys\n" + '\n'.join([f" {k}: {len(k_entries)}" for k, k_entries in duplicates.items()]))
        assert not duplicates
    for entry in entry_dict.values():
        entry.pop('position')
        entry.pop('inline_links')
        entry.pop('resources', None)
    entry_dict = {k: entry_dict[k] for k in sorted(entry_dict)}
    with SEARCH_CACHE.open('wt') as file:
        json.dump(entry_dict, file, indent=4)


def select():
    if SELECTION_PATH.exists():
        with SELECTION_PATH.open('rt') as file:
            selection = json.load(file)
    else:
        selection = {}
    with SEARCH_CACHE.open('rt') as file:
        entries = json.load(file)
    # remove entries
    removed = set(selection).difference(entries)
    if removed:
        print(f"Removing entries that are no longer in search results:\n {', '.join(removed)}\n")
        for key in removed:
            del selection[key]
    # ask for each new entry
    for result_id, entry in entries.items():
        if result_id in selection:
            continue
        elif entry['link'].startswith('https://scholar.archive.org'):
            selection[result_id] = False
            continue
        print(entry['title'])
        print(entry['publication_info']['summary'])
        print(entry['link'])
        response = ask("Add?", {'y': 'Include in bibliography', 'n': 'Skip', 'x': 'Abort selection'})
        if response == 'x':
            break
        elif response == 'y':
            selection[result_id] = True
        elif response == 'n':
            selection[result_id] = False
        else:
            raise RuntimeError(f'{response=}')
    with SELECTION_PATH.open('wt') as file:
        json.dump(selection, file, indent=1)


def download():
    "Download BiBTeX entries"
    if BIBTEX_CACHE.exists():
        with BIBTEX_CACHE.open('rb') as file:
            bibtex_cache = pickle.load(file)
    else:
        bibtex_cache = {}
    with SELECTION_PATH.open('rt') as file:
        selections = json.load(file)
    # download missing citations
    new = 0
    for result_id, include in selections.items():
        if not include or result_id in bibtex_cache:
            continue
        r = requests.get(f'https://serpapi.com/search.json?engine=google_scholar_cite&q={result_id}&api_key={API_KEY}')
        content = json.loads(r.content)
        if 'links' not in content:
            print(f"Links not found:\n{content}")
            break
        for style_entry in content['links']:
            if style_entry['name'] == 'BibTeX':
                bibtex_url = style_entry['link']
                break
        else:
            print(f"BibTeX link not found in:\n{r.content.decode()}")
            break

        # wait_hours = 0
        while True:
            r = requests.get(bibtex_url)
            if r.status_code == 403:
                print(f"Error 403:\n{r.content.decode()}\n")
                # wait_hours += 1
                base = 10 * 60 * 60
                sleep_time = random.uniform(base + 5.432, base + 65.476)
                print(f"sleeping for {sleep_time/3600:.1f} hours")
                time.sleep(sleep_time)
                continue
            break
        print(result_id, end=' - ')
        bibtex_cache[result_id] = r.content
        with BIBTEX_CACHE.open('wb') as file:
            pickle.dump(bibtex_cache, file)
        new += 1
        # if new % 2:
        sleep_time = random.uniform(65.432, 605.476)
        # else:
        #     base = 60*60
        #     sleep_time = random.uniform(base + 5.432, base + 335.476)
        print(f"sleeping for {sleep_time/60:.1f} minutes")
        time.sleep(sleep_time)
    if not new:
        print("No new reference")


def parse():
    with BIBTEX_CACHE.open('rb') as file:
        bibtex_entries = pickle.load(file)  # individual BibTeX entries downloaded from scholar
    with SEARCH_CACHE.open('rt') as file:
        raw_entries = json.load(file)  # initial search on scholar
    with SELECTION_PATH.open('rt') as file:
        selection = json.load(file)
    # Load existing citation data
    bib = parse_file(DST, 'bibtex')
    # Remove entries that have been removed upstream
    removed = [key for key, entry in bib.entries.items() if 'google_result_id' in entry.fields and not selection.get(entry.fields['google_result_id'], False)]
    if removed:
        print(f"Removing {', '.join(removed)}")
    # add new
    unseen_keys = set(bib.entries.keys())
    for result_id, raw_bibtex in bibtex_entries.items():
        if not selection.get(result_id, False):
            continue
        elif result_id in OBSOLETE:
            continue
        raw_bibtex = re.sub(rb'[^\x00-\x7F]+', b' ', raw_bibtex)
        # parse entry
        data = parse_bytes(raw_bibtex, 'bibtex')
        assert len(data.entries) == 1
        ((key, entry),) = data.entries.items()
        if key in bib.entries and key in unseen_keys:
            unseen_keys.remove(key)
            continue
        # For later additions, use Google ID
        elif result_id in bib.entries:
            if result_id in unseen_keys:
                unseen_keys.remove(result_id)
                continue
            raise RuntimeError(f"Duplicate key {result_id}")
        url = raw_entries[result_id]['link']
        journal = entry.fields.get('journal', '').lower()
        doi = None
        if entry.type == 'phdthesis':
            pass
        elif journal == 'biorxiv':
            if match := re.match(r"https://www\.biorxiv\.org/content/([\d./]+)\.abstract", url):
                doi = match.group(1)
            elif match := re.match(r"https://www\.biorxiv\.org/content/biorxiv/early/([\d./]+)\.full.pdf", url):
                pass  # DOI unknown
            else:
                raise RuntimeError(f"Can't identify DOI from {url=}")
        elif not journal:
            if entry.fields.get('publisher') == 'PsyArXiv':
                entry.fields['journal'] = '{PsyArXiv}'
            else:
                entry.fields['journal'] = '???'
                print('Warning: missing journal')
        if doi:
            if doi in BIORXIV_OBSOLETE:
                continue
            entry.fields['doi'] = doi
        else:
            entry.fields['url'] = url
        entry.fields['google_result_id'] = result_id
        # fix title
        for repl in ACRONYMS:
            entry.fields['title'] = re.sub(repl, '{' + repl + '}', entry.fields['title'])
        bib.add_entry(result_id, entry)
    # Fix escape-char bug in PybTex
    out = bib.to_bytes('bibtex')
    for bad_char in [br'\&', b'_']:
        out = out.replace(b'\\' + bad_char, bad_char)
    DST.write_bytes(out)


if __name__ == '__main__':
    parser = ArgumentParser(description="Find publications that used Eelbrain on Google Scholar")
    parser.add_argument('task', choices=('search', 'select', 'download', 'parse'))
    parser.add_argument('--cached', default=False, help="When searching, used cached entries")
    args = parser.parse_args()
    if args.task == 'search':
        search(args.cached)
    elif args.task == 'select':
        select()
    elif args.task == 'download':
        download()
    elif args.task == 'parse':
        parse()



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"Private functionality from mne.channels.channels"
import mne
from mne.io.constants import FIFF


def _adjacency_id(info: mne.Info, ch_type: str):
    "Guess system ID for channel-adjacency; based on :func:`mne.channels.find_ch_adjacency`"
    # Excerpt from mne.channels.find_ch_adjacency
    (has_vv_mag, has_vv_grad, is_old_vv, has_4D_mag, ctf_other_types, has_CTF_grad, n_kit_grads, has_any_meg, has_eeg_coils, has_eeg_coils_and_meg, has_eeg_coils_only, has_neuromag_122_grad, has_csd_coils) = _get_ch_info(info)
    conn_name = None
    if has_vv_mag and ch_type == 'mag':
        conn_name = 'neuromag306mag'
    elif has_vv_grad and ch_type == 'grad':
        conn_name = 'neuromag306planar'
    elif has_4D_mag:
        if any((key := f'MEG {i}') in info['ch_names'] for i in range(149, 249)):
            idx = info['ch_names'].index(key)
            grad = info['chs'][idx]['coil_type'] == FIFF.FIFFV_COIL_MAGNES_GRAD
            mag = info['chs'][idx]['coil_type'] == FIFF.FIFFV_COIL_MAGNES_MAG
            if ch_type == 'grad' and grad:
                conn_name = 'bti248grad'
            elif ch_type == 'mag' and mag:
                conn_name = 'bti248'
        elif 'MEG 148' in info['ch_names'] and ch_type == 'mag':
            idx = info['ch_names'].index('MEG 148')
            if info['chs'][idx]['coil_type'] == FIFF.FIFFV_COIL_MAGNES_MAG:
                conn_name = 'bti148'
    elif has_CTF_grad and ch_type == 'mag':
        if info['nchan'] < 100:
            conn_name = 'ctf64'
        elif info['nchan'] > 200:
            conn_name = 'ctf275'
        else:
            conn_name = 'ctf151'
    # End excerpt
    return conn_name


# copied from mne.channels.channels
def _get_ch_info(info):
    """Get channel info for inferring acquisition device."""
    chs = info['chs']
    # Only take first 16 bits, as higher bits store CTF comp order
    coil_types = {ch['coil_type'] & 0xFFFF for ch in chs}
    channel_types = {ch['kind'] for ch in chs}

    has_vv_mag = any(k in coil_types for k in
                     [FIFF.FIFFV_COIL_VV_MAG_T1, FIFF.FIFFV_COIL_VV_MAG_T2,
                      FIFF.FIFFV_COIL_VV_MAG_T3])
    has_vv_grad = any(k in coil_types for k in [FIFF.FIFFV_COIL_VV_PLANAR_T1,
                                                FIFF.FIFFV_COIL_VV_PLANAR_T2,
                                                FIFF.FIFFV_COIL_VV_PLANAR_T3])
    has_neuromag_122_grad = any(k in coil_types
                                for k in [FIFF.FIFFV_COIL_NM_122])

    is_old_vv = ' ' in chs[0]['ch_name']

    has_4D_mag = FIFF.FIFFV_COIL_MAGNES_MAG in coil_types
    ctf_other_types = (FIFF.FIFFV_COIL_CTF_REF_MAG,
                       FIFF.FIFFV_COIL_CTF_REF_GRAD,
                       FIFF.FIFFV_COIL_CTF_OFFDIAG_REF_GRAD)
    has_CTF_grad = (FIFF.FIFFV_COIL_CTF_GRAD in coil_types or
                    (FIFF.FIFFV_MEG_CH in channel_types and
                     any(k in ctf_other_types for k in coil_types)))
    # hack due to MNE-C bug in IO of CTF
    # only take first 16 bits, as higher bits store CTF comp order
    n_kit_grads = sum(ch['coil_type'] & 0xFFFF == FIFF.FIFFV_COIL_KIT_GRAD
                      for ch in chs)

    has_any_meg = any([has_vv_mag, has_vv_grad, has_4D_mag, has_CTF_grad,
                       n_kit_grads])
    has_eeg_coils = (FIFF.FIFFV_COIL_EEG in coil_types and
                     FIFF.FIFFV_EEG_CH in channel_types)
    has_eeg_coils_and_meg = has_eeg_coils and has_any_meg
    has_eeg_coils_only = has_eeg_coils and not has_any_meg
    has_csd_coils = (FIFF.FIFFV_COIL_EEG_CSD in coil_types and
                     FIFF.FIFFV_EEG_CH in channel_types)

    return (has_vv_mag, has_vv_grad, is_old_vv, has_4D_mag, ctf_other_types,
            has_CTF_grad, n_kit_grads, has_any_meg, has_eeg_coils,
            has_eeg_coils_and_meg, has_eeg_coils_only, has_neuromag_122_grad,
            has_csd_coils)



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: sphinx
#       format_version: '1.1'
#       jupytext_version: 1.10.2
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

"""
.. _exa-utsstat:

UTS-statistics (plot.UTSStat)
=============================

"""
# sphinx_gallery_thumbnail_number = 5
from eelbrain import *


ds = datasets.simulate_erp(snr=1)
# add an orthogonal variable
ds['length'] = Factor(ds['n_chars'] > 4, labels={True: 'long', False: 'short'})
# smooth the data to make plots nicer
ds['cz'] = ds["eeg"].sub(sensor='Cz').smooth('time', 0.05)

###############################################################################
# Plot the mean and SEM by condition:

p = plot.UTSStat('cz', 'predictability', data=ds, axh=3)

###############################################################################
# Split between different axes and use a ``t``-frame:

p = plot.UTSStat('cz', 'predictability', 'length', data=ds, frame='t', axh=2)

###############################################################################
# Colors can also be set using the within-plot category only:

colors = plot.colors_for_oneway(['high', 'low'])
p = plot.UTSStat("cz", 'predictability', 'length', data=ds, colors=colors, frame='t', legend=False, axh=2)
legend = plot.ColorList(colors)

###############################################################################
# Plot all categories in the same plot. Make lines more distinguishable through line-style:

colors = {
    ('high', 'long'): plot.Style((1, 0, 0)),
    ('high', 'short'): plot.Style((1, 0.5, 0.5), linestyle='--'),
    ('low', 'long'): plot.Style((0, 0, 1)),
    ('low', 'short'): plot.Style((0.5, 0.5, 1), linestyle='--'),
}

p = plot.UTSStat("cz", 'predictability % length', data=ds, colors=colors, frame='t', top=1e-6, axh=3)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
# This script can be executed from the Terminal using $ pythonw
# dataset: mne_sample
from eelbrain import *


w = 3

# load sample data
ds = datasets.get_mne_sample(sub="modality=='A'", sns=True)


# interactive plot
plot.TopoButterfly('meg', data=ds)


# plot topography at a specific time point
plot.Topomap("meg.sub(time=.07)", data=ds)

# plot topography at a specific time point by condition
plot.Topomap("meg.sub(time=.07)", 'side', data=ds)

# plot topography at a specific time point for one condition;
# clip with a circular outline
plot.Topomap("meg.sub(side=='L', time=.07)", data=ds, clip='circle')

# plot topography with a specific projection
plot.Topomap("meg.sub(time=.07)", proj='back', data=ds)

# same but without clipping the map
plot.Topomap("meg.sub(time=.07)", proj='back', clip=False, data=ds)


# mark sensors (all sensors with z coordinate < 0
y = ds['meg'].sub(time=.07).mean('case')
mark = ds['meg'].sensor.z < 0
plot.Topomap(y, mark=mark)

# different cmap, and manually adjusted head outline
plot.Topomap(y, head_radius=(0.3, 0.33), head_pos=0.02, cmap='jet')

# sensor labels
plot.SensorMap(y, mark=mark, head_radius=(0.3, 0.33), head_pos=0.02,
               labels='name')

# sensor map with labels
p = plot.SensorMap(y, mark=mark, w=9, head_radius=(0.3, 0.33), head_pos=0.02)
# move overlapping labels apart
p.separate_labels()


# run the GUI if the script is executed from the shell
if __name__ == '__main__':
    gui.run()



"""Use r (rpy2) for testing"""
import warnings
from rpy2.robjects import r

try:
    from rpy2.rinterface import RRuntimeWarning
except ImportError:  # rpy2 < 2.8
    RRuntimeWarning = UserWarning


def r_require(package):
    with r_warning_filter:
        success = r('require(%s)' % package)[0]

    if not success:
        print(r("install.packages('%s', repos='http://cran.us.r-project.org')"
                % package))
        success = r('require(%s)' % package)[0]
        if not success:
            raise RuntimeError("Could not install R package %r" % package)


class RWarningFilter:

    def __enter__(self):
        self.context = warnings.catch_warnings()
        self.context.__enter__()
        warnings.filterwarnings('ignore', category=RRuntimeWarning)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.context.__exit__(exc_type, exc_val, exc_tb)


r_warning_filter = RWarningFilter()



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: sphinx
#       format_version: '1.1'
#       jupytext_version: 1.5.0
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

"""
.. _exa-colors:

Colors
======

In general, colors for categorial variables are set as ``{cell: color}`` dictionaries. For factors, cells are strings, and for interactions they are tuples of strings. Colors can be set to any color understood by matplotlib.
"""
# sphinx_gallery_thumbnail_number = 3
from eelbrain import *


ds = datasets.get_uv()

colors = {'a1': (1, 0, 0), 'a2': (0, 0, 1)}
p = plot.Barplot('fltvar', 'A', data=ds, w=2, colors=colors)

colors = {
    ('a1', 'b1'): 'red', 
    ('a1', 'b2'): (1, 1, 0),
    ('a2', 'b1'): (0, 0, 1),
    ('a2', 'b2'): (0, 1, 1),
}
p = plot.Barplot('fltvar - 0.5', 'A % B', data=ds, w=2, colors=colors)

###############################################################################
# Unambiguous colors
# ^^^^^^^^^^^^^^^^^^
# Eelbrain also comes with utilities to quickly generate such color dictionaries. `Unambiguous colors <https://jfly.uni-koeln.de/html/color_blind/#pallet>`_ are colors for up to 8 categories that are universally distinguishable.

# add a Factor with more categories
ds['AB'] = ds.eval('A%B').as_factor()

colors = plot.colors_for_oneway(ds['AB'].cells, unambiguous=True)
p = plot.Barplot('fltvar - 0.5', 'AB', data=ds, w=2, xticks=False, colors=colors)
p_colors = plot.ColorList(colors)

colors = plot.colors_for_oneway(ds['AB'].cells, unambiguous=[2, 5, 3, 6])
p = plot.Barplot('fltvar - 0.5', 'AB', data=ds, w=2, xticks=False, colors=colors)
p_colors = plot.ColorList(colors)

###############################################################################
# Hue-based colors
# ^^^^^^^^^^^^^^^^
# Colors can also be generated based on a color wheel.

cells = 'abc'
colors = plot.colors_for_oneway(cells)
p = plot.ColorList(colors)

cells = 'abc'
colors = plot.colors_for_oneway(cells, hue_start=0.3)
p = plot.ColorList(colors)

cells = 'abcdefgh'
colors = plot.colors_for_oneway(cells)
p = plot.ColorList(colors)

###############################################################################
# Hue and lightness can be used to distinguish two factors in an interaction.

cells_a = ('a', 'b')
cells_b = ('1', '2')
colors = plot.colors_for_twoway(cells_a, cells_b)
p = plot.ColorGrid(cells_a, cells_b, colors, h=1, w=1)

cells_a = ('a', 'b')
cells_b = ('1', '2', '3')
colors = plot.colors_for_twoway(cells_a, cells_b)
p = plot.ColorGrid(cells_a, cells_b, colors, h=1, w=1)

cells_a = ('a', 'b', 'c')
cells_b = ('1', '2')
colors = plot.colors_for_twoway(cells_a, cells_b, hue_start=0.3)
p = plot.ColorGrid(cells_a, cells_b, colors, h=1, w=1)



"""
.. _exa-dataset:

Dataset basics
==============

.. contents:: Contents
   :local:

Load and prepare an example dataset:
"""
# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from eelbrain import *
import numpy
import pandas


df = pandas.read_csv('https://vincentarelbundock.github.io/Rdatasets/csv/psych/Tal_Or.csv')
data = Dataset.from_dataframe(df)
data['cond'] = data['cond'].as_factor({0: 'low', 1: 'high'})
data['gender'] = data['gender'].as_factor({1: 'male', 2: 'female'})

###############################################################################
# Inspecting datasets
# -------------------
#
# The whole dataset can be displayed like any variable in iPython (in a plain text environment, use ``print(data)``).
# For larger datasets it can be more convenient to print only the first few cases...

data.head()

###############################################################################
# ... or a summary of variables:
data.summary()

###############################################################################
# Individual rows and columns can be retrieved with common indexing: 

data[10:15]

###############################################################################

data[2]

###############################################################################

data['age']

###############################################################################
# Using datasets in functions
# ---------------------------
#
# Datasets collect information describing the same cases (rows) on different variables (columns).
# This can simplify calling functions that combine information from multiple columns.
# Columns can be supplied as strings, and the dataset in the ``data`` parameter:

table.frequencies('cond', 'gender', data=data)

###############################################################################

p = plot.Scatter('pmi', 'age', 'gender', data=data, w=3, legend=(.65, .2), alpha=.4)

###############################################################################
# These strings cannot only be keys, but they can be Python code that can be evaluated in the dataset.
# For example, if this is possible:

data.eval('age < 40')  # equivalent to `data['age'] < 40`

###############################################################################
# Then, this can be used directly for plotting:

p = plot.Scatter('pmi', 'age', 'gender', sub="age < 40", data=data, w=3, legend=(.65, .4), alpha=.4)

###############################################################################
# As in other cases, ``%`` is used to specify interaction between categorial variables:

p = plot.Barplot('age', 'cond % gender', data=data, w=3)

###############################################################################
# And ``*`` expands to main effects plus interaction:

test.ANOVA('age', 'cond * gender', data=data)

###############################################################################
# Constructing datasets
# ---------------------
#
# While datasets can be imported from external data sources, it is also often convenient to store new data in a table on the fly.
#
# A dataset can be constructed column by column, by adding one variable after
# another:

# initialize an empty Dataset:
ds = Dataset()
# numeric values are added as Var object:
ds['y'] = Var(numpy.random.normal(0, 1, 6))
# categorical data as represented in Factors:
ds['a'] = Factor(['a', 'b', 'c'], repeat=2)
# A variable that's equal in all cases can be assigned quickly:
ds[:, 'z'] = 0.
# check the result:
ds

###############################################################################
# An alternative way of constructing a dataset is case by case (i.e., row by
# row):

rows = []
for i in range(6):
    subject = f'S{i}'
    y = numpy.random.normal(0, 1)
    a = 'abc'[i % 3]
    rows.append([subject, y, a])
ds = Dataset.from_caselist(['subject', 'y', 'a'], rows, random='subject')
ds



# Author: Proloy Das <proloy@umd.edu>
# cython: boundscheck=False, wraparound=False, language_level=3
# distutils: language = c++
# setuptools: include_dirs = dsyevh3C/
"""
optimized statistics functions

Relies on code from 'Efficient numerical diagonalization of hermitian 3x3 matrices'
governed by GNU LESSER GENERAL PUBLIC LICENSE Version 2.1.
"""
cimport cython
from libc.math cimport sin, cos
cimport numpy as np


cdef double r_TOL = 2.220446049250313e-16

cdef extern from "dsyevh3.c":
    int dsyevh3(double A[3][3], double Q[3][3], double w[3])

cdef extern from "dsytrd3.c":
    dsytrd3(double A[3][3], double Q[3][3], double d[3], double e[2])

cdef extern from "dsyevq3.c":
    int dsyevq3(double A[3][3], double Q[3][3], double w[3])

cdef extern from "dsyevc3.c":
    int dsyevc3(double A[3][3], double w[3])


def rotation_matrices(
        const np.npy_float64[:] phis,
        const np.npy_float64[:] thetas,
        const np.npy_float64[:] xis,
        np.npy_float64[:,:,:] out,
    ):
    cdef unsigned long i, v, case
    cdef double theta, phi, xi
    cdef unsigned long n_cases = phis.shape[0]
    cdef double raxis_x, raxis_y, raxis_z
    cdef double tmp1, tmp2

    cdef double c, s, t

    for case in range(n_cases):
        phi = phis[case]
        theta = thetas[case]
        xi = xis[case]

        raxis_x = sin(phi) * cos(theta)
        raxis_y = sin(phi) * sin(theta)
        raxis_z = cos(phi)
        c = cos(xi)
        s = sin(xi)
        t = 1.0 - c

        out[case, 0, 0] = c + t * raxis_x ** 2
        out[case, 1, 1] = c + t * raxis_y ** 2
        out[case, 2, 2] = c + t * raxis_z ** 2

        tmp1 = raxis_x * raxis_y * t
        tmp2 = raxis_z * s
        out[case, 1, 0] = tmp1 + tmp2
        out[case, 0, 1] = tmp1 - tmp2

        tmp1 = raxis_x * raxis_z * t
        tmp2 = raxis_y * s
        out[case, 2, 0] = tmp1 - tmp2
        out[case, 0, 2] = tmp1 + tmp2

        tmp1 = raxis_y * raxis_z * t
        tmp2 = raxis_x * s
        out[case, 2, 1] = tmp1  + tmp2
        out[case, 1, 2] = tmp1  - tmp2
    return out


@cython.cdivision(True)
def mean_norm_rotated(
        const np.npy_float64[:,:,:] y,
        const np.npy_float64[:,:,:] rotation,
        np.npy_float64[:] out,
):
    cdef unsigned long i, v, case, vi
    cdef double norm, mean

    cdef unsigned long n_cases = y.shape[0]
    cdef unsigned long n_dims = y.shape[1]
    cdef unsigned long n_tests = y.shape[2]

    for i in range(n_tests):
        norm = 0
        for v in range(n_dims):
            mean = 0
            for case in range(n_cases):
                for vi in range(n_dims):
                    mean += rotation[case, v, vi] * y[case, vi, i]
            norm += (mean / n_cases) ** 2
        out[i] = norm ** 0.5
    return out


@cython.cdivision(True)
def t2_stat(
        const np.npy_float64[:,:,:] y,
        np.npy_float64[:] out,
):
    cdef unsigned long i, v, u, case
    cdef double norm, temp, max_eig, TOL

    cdef double mean[3]
    cdef double sigma[3][3]

    cdef double eig[3]
    cdef double vec[3][3]

    cdef unsigned long n_cases = y.shape[0]
    cdef unsigned long n_dims = y.shape[1]
    cdef unsigned long n_tests = y.shape[2]

    for i in range(n_tests):
        # Initialization
        norm = 0
        for v in range(n_dims):
            mean[v] = 0.0
            for u in range(n_dims):
                sigma[u][v] = 0.0
        # Computation
        for case in range(n_cases):
            for v in range(n_dims):
                temp = y[case, v, i]
                mean[v] += temp
                for u in range(n_dims):
                    sigma[u][v] += temp * y[case, u, i]
        for v in range(n_dims):
            for u in range(n_dims):
                sigma[u][v] -= mean[u] * mean[v] / n_cases
        # check non-zero variance
        for v in range(n_dims):
            if sigma[v][v] != 0:
                break
        else:
            out[i] = 0
            continue

        dsyevh3(sigma, vec, eig)
        max_eig = max(eig, 3)
        TOL = r_TOL * max_eig

        for v in range(n_dims):
            temp = 0
            for u in range(n_dims):
                temp += vec[u][v] * mean[u]
            if temp != 0.0:     # Avoid divide by zero
                if eig[v] > TOL:
                    norm += temp ** 2 / eig[v]
                else:
                    norm += temp ** 2 / TOL
        out[i] = norm ** 0.5
    return out


@cython.cdivision(True)
def t2_stat_rotated(
        const np.npy_float64[:,:,:] y,
        const np.npy_float64[:,:,:] rotation,
        np.npy_float64[:] out,
):
    cdef unsigned long i, v, u, case, vi
    cdef double norm, temp, TOL, max_eig

    cdef double mean[3]
    cdef double tempv[3]
    cdef double sigma[3][3]

    cdef double eig[3]
    cdef double vec[3][3]

    cdef unsigned long n_cases = y.shape[0]
    cdef unsigned long n_dims = y.shape[1]
    cdef unsigned long n_tests = y.shape[2]

    for i in range(n_tests):
        # Initialization
        norm = 0
        for v in range(n_dims):
            mean[v] = 0.0
            for u in range(n_dims):
                sigma[u][v] = 0.0
        # Computation
        for case in range(n_cases):
            # rotation
            for u in range(n_dims):
                tempv[u] = 0
                for vi in range(n_dims):
                    tempv[u] += rotation[case, u, vi] * y[case, vi, i]
                mean[u] += tempv[u]
                for v in range(u + 1):      # Only upper triangular part need to be meaningful (See dsyevh.c)
                    sigma[v][u] += tempv[u] * tempv[v]
        for u in range(n_dims):
            for v in range(u + 1):      # Only upper triangular part need to be meaningful (See dsyevh.c)
                sigma[v][u] -= mean[u] * mean[v] / n_cases
        # check non-zero variance
        for v in range(n_dims):
            if sigma[v][v] != 0:
                break
        else:
            out[i] = 0
            continue

        dsyevh3(sigma, vec, eig)
        max_eig = max(eig, 3)
        TOL = r_TOL * max_eig

        for v in range(n_dims):
            temp = 0
            for u in range(n_dims):
                temp += vec[u][v] * mean[u]
            if temp != 0.0:  # Avoid divide by zero
                if eig[v] > TOL:
                    norm += temp ** 2 / eig[v]
                else:
                    norm += temp ** 2 / TOL
        out[i] = norm ** 0.5
    return out


cdef double max(double* x, int n) noexcept:
    cdef int i
    cdef double max_elem = x[0]
    for i in range(1, n):
        if x[i] > max_elem:
            max_elem = x[i]
    return max_elem



from eelbrain import *

configure(False)

ds = datasets.get_mne_sample(-0.1, 0.2, src='ico', sub="modality == 'A'")
res = testnd.TTestOneSample('src', data=ds, tstart=0, tfce=True, samples=100)

# ds = datasets.get_uts()
# res = testnd.TTestOneSample('uts', ds=ds, tfce=True, samples=1000)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
import sys

import numpy as np

from eelbrain import Factor, Var
from eelbrain._stats.permutation import (
    resample, permute_order, permute_sign_flip)


def test_permutation():
    """Test permutation"""
    v = Var(np.arange(6))
    res = np.empty((5, 6))
    for i, y in enumerate(resample(v, samples=5)):
        res[i] = y.x

    # with unit
    s = Factor('abc', tile=2)
    for i, y in enumerate(resample(v, samples=5, unit=s)):
        res[i] = y.x

    # check we have only appropriate cells
    cols = [np.unique(res[:, i]) for i in range(res.shape[1])]
    for i in range(3):
        assert len(np.setdiff1d(cols[i], [i, i + 3])) == 0
    for i in range(3, 6):
        assert len(np.setdiff1d(cols[i], [i, i - 3])) == 0

    # check we have some variability
    assert max(map(len, cols)) == 2

    # make sure sequence is stable
    assert list(map(tuple, permute_order(4, 3))) == [(2, 3, 1, 0), (2, 1, 3, 0), (0, 2, 3, 1)]


def test_permutation_sign_flip():
    "Test permute_sign_flip()"
    res = np.empty((2 ** 6 - 1, 6), dtype=np.int8)
    i = 0
    for i, sign in enumerate(permute_sign_flip(6, samples=-1)):
        res[i] = sign
    assert i == 2 ** 6 - 2
    assert np.all(res.min(1) < 0)  # Check that all permutations have a sign flip
    for i, row in enumerate(res):
        assert not np.any(np.all(row == res[:i], 1))

    # n > 62
    res = list(map(tuple, permute_sign_flip(66, 2)))
    assert len(res[0]) == 66
    assert len(res) == 2
    assert res[0] != res[1]

    # make sure sequence is stable
    if sys.version_info[0] == 3:
        target = [(1, -1, -1, -1), (-1, -1, -1, 1), (-1, 1, -1, -1)]
    else:
        target = [(-1, 1, -1, -1), (-1, -1, 1, -1), (1, -1, -1, 1)]
    assert list(map(tuple, permute_sign_flip(4, 3))) == target



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
import mne

MNE_EPOCHS = mne.BaseEpochs
MNE_RAW = mne.io.BaseRaw
MNE_EVOKED = mne.Evoked
MNE_LABEL = (mne.Label, mne.label.BiHemiLabel)
MNE_VOLUME_STC = (mne.VolSourceEstimate, mne.VolVectorSourceEstimate)



# Eelbrain documentation build configuration file, created by
# sphinx-quickstart on Tue Mar 29 23:16:11 2011.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

from datetime import datetime
from itertools import chain
import os
from pathlib import Path
from warnings import filterwarnings

import eelbrain.plot._brain_object  # make sure that Brain is available
import eelbrain
import mne
from sphinx_gallery.sorting import ExplicitOrder, _SortKey


# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))

autoclass_content = 'class'  # 'both'

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '1.4.3'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.autodoc', 'sphinx.ext.autosummary',  # default
    'sphinx.ext.todo', 'sphinx.ext.imgmath',  # default
    'sphinx.ext.intersphinx',  # http://sphinx.pocoo.org/ext/intersphinx.html
    'sphinx.ext.napoleon',  # https://www.sphinx-doc.org/en/master/usage/extensions/napoleon.html
    'sphinxcontrib.bibtex',  # https://sphinxcontrib-bibtex.readthedocs.io
    'sphinx_gallery.gen_gallery',  # https://sphinx-gallery.github.io
    'sphinx_rtd_theme',  # https://sphinx-rtd-theme.readthedocs.io/
]
# enable to have all methods documented on the same page as a class:
# autodoc_default_flags=['inherited-members']
autosummary_generate = True
autodoc_typehints = "description"

# Napoleon settings
napoleon_google_docstring = False
napoleon_include_init_with_doc = True
napoleon_include_special_with_doc = False
napoleon_use_param = True
napoleon_use_ivar = True
napoleon_use_keyword = True
napoleon_use_rtype = True

bibtex_bibfiles = ['publications.bib']

qualname_overrides = {
    "eelbrain._data_obj.NDVar": "eelbrain.NDVar",
}


################################################################################
# Sphinx-Gallery
################################################################################

def use_pyplot(gallery_conf, fname):
    eelbrain.configure(frame=False)


example_order = {
    'datasets': [
        'intro.py',
        'dataset-basics.py',
        'align.py',
        'ndvar-creating.py',
        'group-level-analysis.py',
    ],
    'statistics': [
        'models.py',
        'RMANOVA.py',
        'ANCOVA.py',
    ],
    'mass-univariate-statistics': [
        'permutation-statistics.py',
        'sensor-ttest.py',
        'sensor-anova.py',
        'sensor-lm.py',
        'sensor-two-stage.py',
        'compare-topographies.py',
    ],
    'plots': [
        'boxplot.py',
        'utsstat.py',
        'colors.py',
        'colormaps.py',
        'customizing.py',
    ],
    'temporal-response-functions': [
        'trf_intro.py',
        'alice-trf.py',
        'mtrf.py',
        'partitions.py',
        'epoch_impulse.py',
    ],
}


class NameOrder(_SortKey):
    """Specify sphinx-gallery example order as file tree"""

    items = list(chain.from_iterable(example_order.values()))

    def __call__(self, item):
        print(f"NameOrder: {item}")
        return self.items.index(item)


sphinx_gallery_conf = {
    'examples_dirs': '../examples',   # path to example scripts
    'subsection_order': ExplicitOrder([f'../examples/{name}' for name in example_order]),
    'within_subsection_order': NameOrder,
    'gallery_dirs': 'auto_examples',  # path where to save gallery generated examples
    'filename_pattern': rf'{os.sep}\w',
    'default_thumb_file': Path(__file__).parent / 'images' / 'eelbrain.png',
    'min_reported_time': 4,
    'download_all_examples': False,
    'reset_modules': ('matplotlib', use_pyplot),
    'reference_url': {'eelbrain': None},
}

# download datasets (to avoid progress bar output in example gallery)
root = mne.datasets.mtrf.data_path()

################################################################################
# Bibliography
################################################################################
# import pybtex.plugin
import pybtex.plugin
from pybtex.style.formatting.unsrt import Style
from pybtex.style.sorting.author_year_title import SortingStyle
from pybtex.style.template import field, href, words


class MySortingStyle(SortingStyle):

    def sorting_key(self, entry):
        author, year, title = SortingStyle.sorting_key(self, entry)
        year = -int(year) if year.isdigit() else 1
        return year, author, title


class MyBibStyle(Style):
    default_sorting_style = MySortingStyle

    def format_url(self, e):
        url = field('url', raw=True)
        return href[url, url]  # will change to href(url)[url]


pybtex.plugin.register_plugin('pybtex.style.formatting', 'year', MyBibStyle)


################################################################################
# Sphinx basics
################################################################################

# Add any paths that contain templates here, relative to this directory.
templates_path = ['templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'Eelbrain'
copyright = u'%i, Christian Brodbeck' % datetime.now().year

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
_version_items = eelbrain.__version__.split('.')
# The short X.Y version.
version = '.'.join(_version_items[:2])
# The full version, including alpha/beta/rc tags.
release = eelbrain.__version__  # '0.0.3'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['build', 'illustrations', 'images', 'static']

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ['eelbrain.']

# Supress warnings for badges
suppress_warnings = ['image.nonlocal_uri']


# -- Custom Options -----------------------------------------------------------

intersphinx_mapping = {
    'python': ('https://docs.python.org/3.8', None),
    'imageio': ('https://imageio.readthedocs.io/en/stable/', None),
    'matplotlib': ('https://matplotlib.org/stable', None),
    'mne': ('https://mne.tools/stable', None),
    'nilearn': ('https://nilearn.github.io/stable', None),
    'numpy': ('https://numpy.org/doc/stable', None),
    'pandas': ('https://pandas.pydata.org/docs', None),
    'rpy2': ('https://rpy2.github.io/doc', None),
    'scipy': ('https://docs.scipy.org/doc/scipy', None),
    'surfer': ('https://pysurfer.github.io', None),
}
# http://sphinx.pocoo.org/ext/intersphinx.html


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
# new default: alabaster
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further. https://sphinx-rtd-theme.readthedocs.io/en/latest/configuring.html
html_theme_options = {
    'logo_only': True,
    'includehidden': False,
    'navigation_depth': 2,  # otherwise RTD includes all autosummary sub-headings
    'html_baseurl': 'https://eelbrain.readthedocs.io/en/stable/',
}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = 'images/eelbrain.png'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = 'static/eelbrain.ico'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['static']
html_css_files = ['css/custom.css']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
html_sidebars = {'**': ['globaltoc.html', 'searchbox.html']}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'Eelbraindoc'


# -- Options for LaTeX output --------------------------------------------------

# The paper size ('letter' or 'a4').
# latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
# latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    ('index', 'Eelbrain.tex', u'Eelbrain Documentation',
     u'Christian Brodbeck', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Additional stuff for the LaTeX preamble.
# latex_preamble = ''

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'eelbrain', u'Eelbrain Documentation',
     [u'Christian Brodbeck'], 1)
]



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
import numpy as np
from numpy.testing import assert_array_equal
import pytest

from eelbrain import Factor, datasets, epoch_impulse_predictor
from eelbrain._trf.shared import DeconvolutionData


def test_deconvolution_data_continuous():
    ds = datasets._get_continuous(ynd=True)

    # normalizing
    data = DeconvolutionData(ds['y'][:8], ds['x1'][:8])
    data.normalize('l1')
    assert data.x[0].mean() == pytest.approx(0, abs=1e-16)
    assert abs(data.x).mean() == pytest.approx(1, abs=1e-10)
    with pytest.raises(ValueError):
        data.apply_basis(0.050, 'hamming')
    # with basis
    data = DeconvolutionData(ds['y'][:8], ds['x1'][:8])
    data.apply_basis(0.500, 'hamming')
    data.normalize('l1')
    assert data.x[0].mean() == pytest.approx(0, abs=1e-16)
    assert abs(data.x).mean() == pytest.approx(1, abs=1e-10)

    # partitioning, no testing set
    data.initialize_cross_validation(4)
    assert len(data.splits.splits) == 4
    assert_array_equal(data.splits.splits[0].validate, [[0, 20]])
    assert_array_equal(data.splits.splits[0].train, [[20, 80]])
    assert_array_equal(data.splits.splits[1].validate, [[20, 40]])
    assert_array_equal(data.splits.splits[1].train, [[0, 20], [40, 80]])
    assert_array_equal(data.splits.splits[2].validate, [[40, 60]])
    assert_array_equal(data.splits.splits[2].train, [[0, 40], [60, 80]])
    assert_array_equal(data.splits.splits[3].validate, [[60, 80]])
    assert_array_equal(data.splits.splits[3].train, [[0, 60]])

    # partitioning, testing set
    data.initialize_cross_validation(4, test=1)
    assert len(data.splits.splits) == 12
    # 0/1
    assert_array_equal(data.splits.splits[0].test, [[0, 20]])
    assert_array_equal(data.splits.splits[0].validate, [[20, 40]])
    assert_array_equal(data.splits.splits[0].train, [[40, 80]])
    # 0/2
    assert_array_equal(data.splits.splits[1].test, [[0, 20]])
    assert_array_equal(data.splits.splits[1].validate, [[40, 60]])
    assert_array_equal(data.splits.splits[1].train, [[20, 40], [60, 80]])
    # 0/3
    assert_array_equal(data.splits.splits[2].test, [[0, 20]])
    assert_array_equal(data.splits.splits[2].validate, [[60, 80]])
    assert_array_equal(data.splits.splits[2].train, [[20, 60]])
    # 1/0
    assert_array_equal(data.splits.splits[3].test, [[20, 40]])
    assert_array_equal(data.splits.splits[3].validate, [[0, 20]])
    assert_array_equal(data.splits.splits[3].train, [[40, 80]])
    # 1/2
    assert_array_equal(data.splits.splits[4].test, [[20, 40]])
    assert_array_equal(data.splits.splits[4].validate, [[40, 60]])
    assert_array_equal(data.splits.splits[4].train, [[0, 20], [60, 80]])
    # 1/3, 2/0, 2/1, 2/3, 3/0, 3/1, 3/2
    assert_array_equal(data.splits.splits[11].test, [[60, 80]])
    assert_array_equal(data.splits.splits[11].validate, [[40, 60]])
    assert_array_equal(data.splits.splits[11].train, [[0, 40]])


def test_deconvolution_data_trials():
    ds = datasets.get_uts()
    n_times = len(ds['uts'].time)
    ds['imp'] = epoch_impulse_predictor('uts', data=ds)
    ds['imp_a'] = epoch_impulse_predictor('uts', "A == 'a1'", data=ds)

    data = DeconvolutionData('uts', 'imp', ds)
    data.normalize('l1')
    assert_array_equal(data.segments, [[i * n_times, (i + 1) * n_times] for i in range(60)])

    # partitioning
    data.initialize_cross_validation(3)
    assert len(data.splits.splits) == 3
    arange = np.arange(len(data.segments)) % 3
    for i, split in enumerate(data.splits.splits):
        validate_index = arange == i
        assert_array_equal(split.validate, data.segments[validate_index])
        assert_array_equal(split.train, data.segments[~validate_index])

    # continuoue model
    data.initialize_cross_validation(3, 'A', ds)
    assert len(data.splits.splits) == 3
    for i, split in enumerate(data.splits.splits):
        validate_index = arange == i
        assert_array_equal(split.validate, data.segments[validate_index])
        assert_array_equal(split.train, data.segments[~validate_index])

    # alternating model
    ds['C'] = Factor('abc', tile=20)
    data.initialize_cross_validation(3, 'C', ds)
    assert len(data.splits.splits) == 3
    arange = np.repeat(np.arange(20), 3) % 3
    for i, split in enumerate(data.splits.splits):
        validate_index = arange == i
        assert_array_equal(split.validate, data.segments[validate_index])
        assert_array_equal(split.train, data.segments[~validate_index])



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
import numpy as np
import scipy.stats
from numpy.testing import assert_allclose
from eelbrain import datasets
from eelbrain._stats import opt
from eelbrain._stats.permutation import permute_sign_flip


def test_t_1samp():
    "Test t_1samp functions"
    ds = datasets.get_uts()
    y = ds.eval("uts.x")
    n_cases = len(y)
    t = np.empty(y.shape[1])

    # t
    opt.t_1samp(y, t)
    t_sp, _ = scipy.stats.ttest_1samp(y, 0)
    assert_allclose(t, t_sp)

    # perm
    t_perm = np.empty_like(t)
    for sign in permute_sign_flip(n_cases, 2):
        opt.t_1samp_perm(y, t_perm, sign)
        opt.t_1samp(y * sign[:, None], t)
        assert_allclose(t_perm, t)



# requires: mayavi
# dataset: mne_sample
'''
Example performes a permutation cluster test on source space data and creates
an HTML file describing he output.
'''
import eelbrain as e


# settings
n_samples = 1000


# create an HTML report in which to document results
report = e.Report("MNE Sample Dataset", author="Prof. Enid Gumby")
section = report.add_section("Introduction")
text = ("A comparison of auditory stimulation to the left vs. the right ear "
        "in the MNE sample dataset. "
        "Spatio-temporal clusters were formed by thresholding the comparison "
        "at a t-value equivalent to an uncorrected p-value of 0.05. For each "
        "cluster, a cluster value was calculated as the sum of all t-values "
        "in the cluster. In order to calculate p-values for each cluster, "
        "a distribution of cluster values was computed by shuffling condition "
        "labels %i times and extracting each time the value of the largest "
        "cluster." % n_samples)
section.append(text)

'''
use the sample dataset loader to load source space data for the mne samples
dataset. Load only auditory data.
'''
ds = e.datasets.get_mne_sample(-0.1, 0.2, src='ico', sub="modality == 'A'")

# Add a table with trial numbers to the report
freqs = e.table.frequencies('side', data=ds)
freq_table = freqs.as_table()
section.add_figure("Number of trials in each condition", freq_table)


'''
perform t-test for side of stimulation (for a more reliable test set samples=10000).
'''
res = e.testnd.TTestIndependent(
    'src', 'side', 'L', 'R', data=ds,
    samples=n_samples,  # number of permutations
    pmin=0.05,  # threshold for clusters (uncorrected p-value)
    tstart=0.05,  # start of the time window of interest
    mintime=0.02,  # minimum duration for clusters
    minsource=10,  # minimum number of sources for clusters
)

# add results to the report
section = report.add_section("Result")

# add an image with all clusters in time bins
pmap = res.masked_parameter_map(1)
image = e.plot.brain.bin_table(pmap, tstep=0.05, surf='smoothwm', views=['lat', 'med'])
section.add_figure("Significant clusters in time bins.", image)
# add a table of all clusters
cluster_table = res.clusters.as_table()
section.add_figure("All clusters", cluster_table)

# plot clusters with p values smaller than 0.25 separately
clusters = res.clusters.sub("p < 0.25")
for i in range(clusters.n_cases):
    p = clusters[i, 'p']

    # add a section with appropriate title
    title = "Cluster %i, p=%.2f" % (i, p)
    subsection = section.add_section(title)

    # retrieve the cluster
    c_i = clusters[i, 'cluster']
    # plot the extent
    c_extent = c_i.sum('time')
    brain = e.plot.brain.cluster(c_extent)
    # add to report (plot.brain plots can be captured using Brain.image())
    image = brain.image("cluster 0 extent")
    brain.close()
    subsection.add_figure("Extent the cluster, p=%s" % p, image)

    # extract and analyze the value in the cluster in each trial
    index = c_i != 0
    c_value = ds['src'].sum(index)
    # index is a boolean NDVar over space and time, so here we are summing in the
    # whole spatio-temporal cluster
    plt_box = e.plot.Boxplot(c_value, 'side', data=ds)
    pw_table = e.test.pairwise(c_value, 'side', data=ds)
    # add to report (create an image from an eelbrain matplotlib plot with .image())
    # use PNG because SVG boxplots do not render correctly in Safari
    image = plt_box.image('image.png')
    section.add_figure("Cluster value", [image, pw_table])

    # extract and analyze the time course in the cluster
    index = c_extent != 0
    c_timecourse = ds['src'].sum(index)
    # c_extent is a boolean NDVar over space only, so here we are summing over the
    # spatial extent of the cluster for every time point but keep the time dimension
    plt_tc = e.plot.UTSStat(c_timecourse, 'side', data=ds)
    # add to report
    image = plt_tc.image()
    section.add_figure("Time course of the average in the largest cluster "
                       "extent", image)

# save the report
report.save_html("Source Permutation Clusters.html")



"""
Export events for use in the Besa pipeline.

Use :func:`meg160_triggers` to export a trigger list for MEG160, then reject
unwanted events, and finally use :func:`besa_evt` to export a corresponding
``.evt`` file.

"""
import numpy as np

from .._data_obj import Var, Dataset
from .._utils import ui
from . import _txt


def meg160_triggers(ds, dest=None, pad=1):
    """
    Export a list of event times used for epoching in MEG-160

    For use together with :func:`besa_evt`. ``save.meg160_triggers(ds)`` adds
    a variable called 'besa_index' which keeps track of the event sequence so
    that the .evt file can be saved correctly after event rejection.

    Parameters
    ----------
    ds : Dataset
        Dataset containing all the desired events.
    dest : None | str
        Path where to save the triggers (if None, a dialog will be displayed).
    pad : int
        Number of epochs to pad with at the beginning and end of the file.
    """
    sfreq = ds.info['raw'].info['sfreq']
    times = ds['i_start'] / sfreq

    # destination
    if dest is None:
        dest = ui.ask_saveas("Save MEG-160 Triggers", "Please pick a "
                             "destination for the MEG-160 triggers",
                             [('MEG160 trigger-list', '*.txt')])
        if not dest:
            return

    # make trigger list with padding
    a = np.ones(len(times) + pad * 2) * times[0]
    a[pad:-pad] = times.x
    triggers = Var(a, name='triggers')

    # export trigger list
    _txt.txt(triggers, dest=dest)

    # index the datafile to keep track of rejections
    ds.index('besa_index', pad)


def besa_evt(ds, tstart=-0.1, tstop=0.6, pad=0.1, dest=None):
    """
    Export an evt file for use with Besa

    For use together with :func:`meg160_triggers`

    Parameters
    ----------
    ds : Dataset
        Dataset containing the events.
    tstart, tstop : scalar [sec]
        start and stop of the actual epoch.
    pad : scalar [sec]
        Time interval added before and after every epoch for padding.
    dest : None | str
        Destination file name. If None, a save as dialog will be shown.

    Notes
    -----
    ds needs to be the same Dataset (i.e., same length) from which the MEG-160
    triggers were created.

    tstart, tstop and pad need to be the same values used for epoching in
    MEG-160.
    """
    idx = ds['besa_index']
    n = idx.x.max() + 1

    # save trigger times in ds
    tstart2 = tstart - pad
    tstop2 = tstop + pad
    epoch_len = tstop2 - tstart2
    start = -tstart2
    stop = epoch_len * n

    # BESA events
    evts = Dataset()
    evts['Tsec'] = Var(np.arange(start, stop, epoch_len))
    evts['Code'] = Var(np.ones(n))

    # remove rejected trials
    evts = evts.sub(idx)

    evts['TriNo'] = Var(ds['trigger'].x)

    # destination
    if dest is None:
        dest = ui.ask_saveas("Save Besa Events", "Please pick a "
                             "destination for the Besa events",
                             [('Besa Events', '*.evt')])
        if not dest:
            return

    evts.save_txt(dest)



from functools import reduce

from mne import SourceSpaces
import numpy as np

from .._data_obj import VolumeSourceSpace, _point_graph
from .._types import PathArg


def merge_volume_source_space(sss, name='Volume'):
    """Merge multiple volume source spaces into one

    Notes
    -----
    Only intended to work properly if ``sss`` is the output of a single
    :func:`mne.setup_volume_source_space` call.
    """
    if len(sss) <= 1:
        return sss
    assert all(ss['type'] == 'vol' for ss in sss)
    ss = sss[0].copy()
    ss['vertno'] = np.unique(np.concatenate([ss['vertno'] for ss in sss]))
    ss['inuse'] = reduce(np.logical_or, (ss['inuse'] for ss in sss))
    ss['nuse'] = ss['inuse'].sum()
    ss['seg_name'] = str(name)
    del ss['neighbor_vert']
    return SourceSpaces([ss], sss.info)


def prune_volume_source_space(
        sss: SourceSpaces,
        grade: int,
        min_neighbors: int = 1,  # remove voxels with fewer neighbors
        mirror: bool = True,  # make source space symmetric
        remove_midline: bool = False,  # remove voxels along the midline
        fill_holes: int = 0,  # fill holes bordered by n voxels
):
    """Remove sources that have ``n`` or fewer neighboring sources"""
    assert len(sss) == 1
    # make copy
    ss = sss[0].copy()
    ss['inuse'] = ss['inuse'].copy()
    assert ss['type'] == 'vol'
    if 'neighbor_vert' in ss:
        del ss['neighbor_vert']
    # coordinates in mm
    rr_mm = np.round(ss['rr'] * 1000).astype(int)
    rr_vertno = {coord: v for v, coord in enumerate(map(tuple, rr_mm))}
    # remove midline
    if remove_midline:
        rm = ss['rr'][ss['vertno']][:, 0] == 0
        ss['inuse'][ss['vertno'][rm]] = 0
        ss['vertno'] = ss['vertno'][~rm]
    # make symmetric
    if mirror:
        vertex_list = list(ss['vertno'])
        for v in ss['vertno']:
            r, a, s = rr_mm[v]
            v_mirror = rr_vertno[-r, a, s]
            if v_mirror not in vertex_list:
                vertex_list.append(v_mirror)
        vertex_list.sort()
        ss['vertno'] = np.array(vertex_list, dtype=ss['vertno'].dtype)
        ss['inuse'][ss['vertno']] = True
    # fill in holes
    if fill_holes:
        assert 6 >= fill_holes >= 3
        max_coord = rr_mm.max(0)
        min_coord = rr_mm.min(0)
        while True:
            add = np.zeros(ss['np'], bool)
            for i in range(ss['np']):
                if ss['inuse'][i]:
                    continue
                ras = rr_mm[i]
                n_neighbors = 0
                for j in range(3):
                    index = list(ras)
                    # up
                    if ras[j] < max_coord[j]:
                        index[j] = ras[j] + grade
                        n_neighbors += rr_vertno[tuple(index)] in ss['vertno']
                    # down
                    if ras[j] > min_coord[j]:
                        index[j] = ras[j] - grade
                        n_neighbors += rr_vertno[tuple(index)] in ss['vertno']
                    # move on if done
                    if n_neighbors >= fill_holes:
                        add[i] = True
                        break
            if not np.any(add):
                break
            ss['inuse'][add] = True
            ss['vertno'] = np.flatnonzero(ss['inuse'])
        ss['nuse'] = len(ss['vertno'])
    # remove points with few neighbors
    if min_neighbors > 1:
        while True:
            neighbors = _point_graph(rr_mm[ss['vertno']], grade * 1.1)
            unique, counts = np.unique(neighbors, return_counts=True)
            keep = unique[counts >= min_neighbors]
            if len(keep) == ss['nuse']:
                break
            ss['vertno'] = ss['vertno'][keep]
            ss['nuse'] = len(ss['vertno'])
            break
        ss['inuse'] = np.isin(np.arange(ss['np']), ss['vertno'])
    # check result
    assert ss['inuse'].sum() == ss['nuse']
    assert np.all(np.flatnonzero(ss['inuse']) == ss['vertno'])
    return SourceSpaces([ss], sss.info)


def restrict_volume_source_space(
        sss: SourceSpaces,
        grade: int,
        subjects_dir: PathArg,
        subject: str,
        parc: str = 'aseg',  # parcellation with labels
        label: str = '*Cortex',  # pattern for selecting labels
        grow: int = 0,  # grow index after selection
):
    assert len(sss) == 1
    # make copy
    ss = sss[0].copy()
    assert ss['type'] == 'vol'
    if 'neighbor_vert' in ss:
        del ss['neighbor_vert']
    # parcellation
    parc = VolumeSourceSpace._read_volume_parc(subjects_dir, subject, parc, ss['rr'][ss['vertno']])
    index = parc.matches(label)
    neighbors = _point_graph(ss['rr'][ss['vertno']], grade * 0.0011)
    neighbors = np.vstack((neighbors, neighbors[:, [1, 0]]))
    for _ in range(grow):
        for i in np.flatnonzero(index):
            index[neighbors[neighbors[:, 0] == i, 1]] = True
    # update ss
    ss['vertno'] = ss['vertno'][index]
    ss['inuse'] = np.isin(np.arange(ss['np']), ss['vertno'])
    ss['nuse'] = len(ss['vertno'])
    return SourceSpaces([ss], sss.info)



from eelbrain import *

# uncomment the following line to embed plots as PNG instead of the default SVG
# configure(format='png')


# create the initial report
report = Report("Test Report", "Test Author")

# add a new section to the report and add some text
s1 = report.add_section("The Big Analysis")
s1.append("This is an introductory paragraph adding some text. It's important "
          "to have some text. What follows is a table showing the first ten "
          "rows of the data, so that it may be examined thoroughly.")

# Add a table to the section
ds = datasets.get_uv()
t = ds.as_table(cases=10, midrule=True, caption="The first ten rows of the "
                                                "data table.")
s1.append(t)

# add a subsection with more text and another table
s11 = s1.add_section("Averages")
s11.append("This is a subordinate section which will show the average of the "
           "data.")

ds_ = ds.aggregate("A%B", drop=['rm'])
t = ds_.as_table(cases=0, midrule=True, caption="Averages of the data.")
s11.append(t)

# add a second subsection with a figure
s12 = s1.add_section("Figure")
s12.append("And now finally we will show a figure")
p = plot.Boxplot('fltvar', 'A%B', 'rm', data=ds)
image = report.add_image_figure(p, "Boxplot of all the data")
p.close()

# add a more complex figure with legend
ds = datasets.get_uts()
p = plot.UTSStat('uts', 'A%B', match='rm', data=ds, legend=False)
p_legend = p.plot_legend()
report.add_figure("Two plots in one figure", [p, p_legend])
p.close()
p_legend.close()

# PNG figure (vector plots with many lines can be slow to display)
p = plot.UTSStat('uts', 'A%B', match='rm', data=ds, error='all')
report.add_figure("All trials", p.image('all', 'png'))
p.close()

# add information on package versions used
report.sign()

# print string representation and save
print(report.get_str())
report.save_html('test_report')



import cProfile
from optparse import OptionParser
import pstats

import mne
from eelbrain import *
import eelbrain

mne.set_log_level('warning')
configure(n_workers=False)


# option parser
parser = OptionParser()
parser.add_option("-m", "--make", dest="make", metavar="KIND",
                  help="Make a new profile of kind mne or uts")
parser.add_option("-f", "--file", dest="file_ext", metavar="NAME",
                  help="Use a different file for this profile")
parser.add_option("-s", "--sort", dest="sort", metavar="CRITERION",
                  help="Sort the profile entries according to CRITERION")
parser.add_option("-n", dest="number", metavar="NUMBER",
                  help="Display NUMBER entries from the profile.")
(options, args) = parser.parse_args()


# process options
if options.file_ext is None:
    fname = 'profile_of_anova.profile'
else:
    fname = 'profile_of_anova_%s.profile' % options.file_ext

sort = options.sort
if options.number is None:
    number = 20
    if sort is None:
        sort = 'tottime'
else:
    number = int(options.number)


# run profile
make = options.make
if make:
    if make == 'mne':
        ds = datasets.get_mne_sample(-0.1, 0.2, src='ico', sub=None, rm=True)
        statement = ("testnd.ANOVA('src', 'side * modality * subject', "
                     "match='subject', ds=ds, samples=2, tstart=0)")
    elif make == 'uts':
        ds = datasets.get_rand(True)
        statement = ("testnd.ANOVA('uts', 'A * B * rm', ds=ds, samples=100, "
                     "tfce=True, tstart=0, match='rm')")
    else:
        raise ValueError("-m %s" % make)
    cProfile.run(statement, fname)


# display stats
p = pstats.Stats(fname)
if sort:
    p.sort_stats(sort)
p.print_stats(number)



"""Some basic example datasets for testing."""
from itertools import product
from pathlib import Path
import shutil
import string
from typing import Literal, Optional, Tuple, Union
from os.path import join

import mne
from mne import minimum_norm as mn
import numpy as np
from mne_bids import BIDSPath, write_raw_bids

from .. import _info, load
from .._data_obj import Dataset, Factor, Var, NDVar, Case, Categorial, Scalar, Sensor, Space, UTS
from .._ndvar import concatenate, convolve
from .._types import PathArg


def permute(variables):
    """Create a Dataset from permuting variables.

    Parameters
    ----------
    variables : sequence
        Sequence of (name, values) tuples. For examples:
        ``(('A', ('a1', 'a2')), ('B', ('b1', 'b2')))``

    Examples
    --------
    >>> print(permute((('A', ('a1', 'a2')),('B', ('b1', 'b2')))))
    A    B
    -------
    a1   b1
    a1   b2
    a2   b1
    a2   b2
    """
    names = (v[0] for v in variables)
    cases = tuple(product(*(v[1] for v in variables)))
    return Dataset.from_caselist(names, cases)


def _get_continuous(
        n_samples: int = 100,
        ynd: bool = False,
        seed: int = 0,
        xn: int = 0,
) -> Dataset:
    """Generate continuous data for temporal response function tests

    Parameters
    ----------
    n_samples
        Number of samples to simulate.
    ynd
        Include 3d ``y``.
    seed
        Seed for :func:`numpy.random.seed` (``None`` to skip seeding random
        state; default us 0).
    xn
        Number of rows in ``xn``.

    Returns
    -------
    data
        Dataset with simulated data:

         - ``x1``: random time series
         - ``x2``: two random time series
         - ``h1`` and ``h2``: Kernels corresponding to ``x1`` and ``x2``
         - ``y``: convolution of ``(x1 * h1) + (x2 * h2)``
    """
    random = np.random if seed is None else np.random.RandomState(seed)
    time = UTS(0, 0.1, n_samples)
    h_time = UTS(0, 0.1, 10)
    xdim = Scalar('xdim', [0, 1])

    x1 = NDVar(random.normal(0, 1, (n_samples,)), (time,), 'x1')
    h1 = NDVar(np.array([0, 0, 1, 3, 0, 0, 0, 0, 2, 3]), (h_time,), 'h1')

    x2 = NDVar(random.normal(0, 1, (2, n_samples,)), (xdim, time), 'x2')
    h2 = NDVar(np.array([[0, 0, 0, 0, 0, 0, -1, -3, 0, 0],
                         [0, 0, 2, 2, 0, 0, 0, 0, 0, 0]]),
               (xdim, h_time), 'h2')

    y = convolve(h1, x1)
    y += convolve(h2, x2)
    y.name = 'y'
    y.info = _info.for_eeg()
    out = {'y': y, 'x1': x1, 'h1': h1, 'x2': x2, 'h2': h2}
    if ynd:
        dim = Sensor([[-1, 0, 0], [1, 0, 0]], ['SL', 'SR'], 'TEST-2-CH')
        out['h1nd'] = h1 = concatenate([h1, -h1], dim, 'h1')
        out['h2nd'] = h2 = concatenate([h2, h2 * 0], dim, 'h2')
        out['ynd'] = convolve(h1, x1) + convolve(h2, x2)
    if xn:
        xn_dim = Scalar('xdim', range(xn))
        out['xn'] = NDVar(random.normal(0, 1, (xn, n_samples)), (xn_dim, time))
        out['hn'] = NDVar(random.normal(0, 1, (xn, 10)), (xn_dim, h_time))
        out['yn'] = convolve(out['hn'], out['xn'])
        out['yn'] += random.normal(0, 1, (n_samples,))
    return Dataset(out)


def get_loftus_masson_1994():
    "Dataset used for illustration purposes by Loftus and Masson (1994)"
    ds = Dataset()
    ds['subject'] = Factor(range(1, 11), tile=3, random=True)
    ds['exposure'] = Var([1, 2, 5], repeat=10)
    ds['n_recalled'] = Var([10, 6, 11, 22, 16, 15, 1, 12, 9, 8,
                            13, 8, 14, 23, 18, 17, 1, 15, 12, 9,
                            13, 8, 14, 25, 20, 17, 4, 17, 12, 12])
    return ds


def get_mne_epochs():
    """MNE-Python Epochs"""
    data_path = Path(mne.datasets.sample.data_path())
    raw_path = data_path / 'MEG' / 'sample' / 'sample_audvis_raw.fif'
    events_path = data_path / 'MEG' / 'sample' / 'sample_audvis_raw-eve.fif'
    raw = mne.io.Raw(raw_path)
    events = mne.read_events(events_path)
    epochs = mne.Epochs(raw, events, 32, -0.1, 0.4, preload=True)
    return epochs


def get_mne_evoked(ndvar=False):
    """MNE-Python Evoked

    Parameters
    ----------
    ndvar : bool
        Convert to NDVar (default False).
    """
    data_path = Path(mne.datasets.sample.data_path())
    evoked_path = data_path / 'MEG' / 'sample' / 'sample_audvis-ave.fif'
    evoked = mne.Evoked(evoked_path, "Left Auditory")
    if ndvar:
        return load.mne.evoked_ndvar(evoked)
    else:
        return evoked


def get_mne_stc(ndvar=False, src='ico-5', subject='sample'):
    """MNE-Python SourceEstimate

    Parameters
    ----------
    ndvar : bool
        Convert to NDVar (default False; src="ico-4" is false, but it works as
        long as the source space is not accessed).
    src : 'ico-5' | 'vol-7' | 'oct-4
        Source space to use.

    Notes
    -----
    Source space only available for ``oct-4``, ``sample``.
    """
    data_path = Path(mne.datasets.testing.data_path())
    meg_sdir = data_path / 'MEG' / 'sample'
    subjects_dir = data_path / 'subjects'
    # scaled subject
    if subject == 'fsaverage_scaled':
        subject_dir = subjects_dir / subject
        if not subject_dir.exists():
            mne.scale_mri('fsaverage', subject, .9, subjects_dir=subjects_dir, skip_fiducials=True, labels=False, annot=True)
        data_subject = 'fsaverage'
    else:
        data_subject = subject

    if src == 'vol-7':
        inv = mn.read_inverse_operator(str(meg_sdir / 'sample_audvis_trunc-meg-vol-7-meg-inv.fif'))
        evoked = mne.read_evokeds(str(meg_sdir / 'sample_audvis_trunc-ave.fif'), 'Left Auditory')
        stc = mn.apply_inverse(evoked, inv, method='MNE', pick_ori='vector')
        if data_subject == 'fsaverage':
            m = mne.compute_source_morph(stc, 'sample', data_subject, subjects_dir)
            stc = m.apply(stc)
            stc.subject = subject
        elif subject != 'sample':
            raise ValueError(f"subject={subject!r}")
        if ndvar:
            return load.mne.stc_ndvar(stc, subject, 'vol-7', subjects_dir, 'MNE', sss_filename='{subject}-volume-7mm-src.fif')
        else:
            return stc
    elif src == 'oct-4':
        if subject != 'sample':
            raise ValueError(f"subject={subject!r}: source space only available for 'sample'")
        inv = mne.minimum_norm.read_inverse_operator(meg_sdir / 'sample_audvis_trunc-meg-eeg-oct-4-meg-inv.fif')
        evokeds = mne.read_evokeds(meg_sdir / 'sample_audvis_trunc-ave.fif')
        evoked = mne.combine_evoked([evokeds[i].apply_baseline() for i in [0, 1]], [1, 1])
        stc = mne.minimum_norm.apply_inverse(evoked, inv)
    elif src == 'ico-5':
        stc_path = meg_sdir / f'{data_subject}_audvis_trunc-meg'
        stc = mne.read_source_estimate(str(stc_path), subject)
    else:
        raise ValueError(f"src={src!r}")

    if ndvar:
        return load.mne.stc_ndvar(stc, subject, src, subjects_dir)
    else:
        return stc


def _mne_source_space(subject, src_tag, subjects_dir):
    """Load mne source space

    Parameters
    ----------
    subject : str
        Subejct
    src_tag : str
        Spacing (e.g., 'ico-4').
    """
    src_file = Path(subjects_dir) / subject / 'bem' / f'{subject}-{src_tag}-src.fif'
    src, spacing = src_tag.split('-')
    if src_file.exists():
        return mne.read_source_spaces(src_file, False)
    elif src == 'ico':
        ss = mne.setup_source_space(subject, spacing=src + spacing, subjects_dir=subjects_dir, add_dist=True)
    elif src == 'vol':
        mri_file = subjects_dir / subject / 'mri' / 'orig.mgz'
        bem_file = subjects_dir / subject / 'bem' / 'sample-5120-5120-5120-bem-sol.fif'
        ss = mne.setup_volume_source_space(subject, pos=float(spacing), mri=mri_file, bem=bem_file, mindist=0., exclude=0., subjects_dir=subjects_dir)
    else:
        raise ValueError(f"{src_tag=}")
    mne.write_source_spaces(src_file, ss)
    return ss


def get_mne_sample(
        tmin: float = -0.1,
        tmax: float = 0.4,
        baseline: Optional[Tuple[Optional[float], Optional[float]]] = (None, 0),
        sns: Union[bool, Literal['eeg', 'mag', 'grad']] = False,
        src: Literal[False, 'ico', 'vol'] = None,
        sub: str = "modality=='A'",
        ori: Literal['free', 'fixed', 'vector'] = 'free',
        snr: float = 2,
        method: str = 'dSPM',
        rm: bool = False,
        stc: bool = False,
        hpf: float = 0,
        proj: bool = True,
        fsaverage: bool = False,
):
    """Load events and epochs from the MNE sample data

    Parameters
    ----------
    tmin
        Relative time of the first sample of the epoch.
    tmax
        Relative time of the last sample of the epoch.
    baseline
        Period for baseline correction.
    sns
        Add sensor space data as NDVar as ``ds['meg']`` (default ``False``).
        Set to ``'grad'`` to load gradiometer data.
    src
        Add source space data as NDVar as ``ds['src']`` (default ``False``).
    sub
        Expression for subset of events to load. For a very small dataset use e.g.
        ``[0,1]``.
    ori
        Orientation of sources.
    snr
        MNE inverse parameter.
    method
        MNE inverse parameter.
    rm
        Pretend to be a repeated measures dataset (adds 'subject' variable).
    stc
        Add mne SourceEstimate for source space data as ``ds['stc']`` (default
        ``False``).
    hpf
        High pass filter cutoff.
    proj
        Add projectors.
    fsaverage
        Morph data to FSAverage template brain.

    Returns
    -------
    ds : Dataset
        Dataset with epochs from the MNE sample dataset in ``ds['epochs']``.
    """
    if ori == 'free':
        loose = 1
        fixed = False
        pick_ori = None
    elif ori == 'fixed':
        loose = 0
        fixed = True
        pick_ori = None
    elif ori == 'vector':
        loose = 1
        fixed = False
        pick_ori = 'vector'
    else:
        raise ValueError(f"{ori=}")

    data_dir = Path(mne.datasets.sample.data_path())
    meg_dir = data_dir / 'MEG' / 'sample'
    raw_file = meg_dir / 'sample_audvis_filt-0-40_raw.fif'
    event_file = meg_dir / 'sample_audvis_filt-0-40-eve.fif'
    subjects_dir = data_dir / 'subjects'
    subject = 'sample'
    label_path = subjects_dir / subject / 'label' / '%s.label'

    if not event_file.exists():
        raw = mne.io.Raw(raw_file)
        events = mne.find_events(raw, stim_channel='STI 014')
        mne.write_events(event_file, events)
    ds = load.mne.events(raw_file, events=event_file)
    if hpf:
        ds.info['raw'].load_data()
        ds.info['raw'].filter(hpf, None, n_jobs=1)
    ds.index()
    ds.info['subjects_dir'] = subjects_dir
    ds.info['subject'] = subject
    ds.info['label'] = label_path

    # get the trigger variable form the dataset for eaier access
    trigger = ds['trigger']

    # use trigger to add various labels to the dataset
    ds['condition'] = Factor(trigger, labels={1: 'LA', 2: 'RA', 3: 'LV', 4: 'RV', 5: 'smiley', 32: 'button'})
    ds['side'] = Factor(trigger, labels={1: 'L', 2: 'R', 3: 'L', 4: 'R', 5: 'None', 32: 'None'})
    ds['modality'] = Factor(trigger, labels={1: 'A', 2: 'A', 3: 'V', 4: 'V', 5: 'None', 32: 'None'})

    if rm:
        ds = ds.sub('trigger < 5')
        ds = ds.equalize_counts('side % modality')
        subject_f = ds.eval('side % modality').enumerate_cells()
        ds['subject'] = subject_f.as_factor('s%r', random=True)

    if sub:
        ds = ds.sub(sub)

    load.mne.add_mne_epochs(ds, tmin, tmax, baseline)
    if sns:
        data_arg = 'mag' if sns is True else sns
        ds['meg'] = load.mne.epochs_ndvar(ds['epochs'], data=data_arg, sysname='neuromag306', proj=proj)

    if not src:
        return ds
    elif src == 'ico':
        src_tag = inv_tag = 'ico-4'
        if fixed:
            inv_tag += '-fixed'
    elif src == 'vol':
        src_tag = inv_tag = 'vol-10'
    else:
        raise ValueError(f"{src=}")
    epochs = ds['epochs']

    # get inverse operator
    inv_file = meg_dir / f'sample_eelbrain_{inv_tag}-inv.fif'
    if inv_file.exists():
        inv = mne.minimum_norm.read_inverse_operator(inv_file)
    else:
        fwd_file = meg_dir / f'sample-{src_tag}-fwd.fif'
        bem_dir = subjects_dir / subject / 'bem'
        bem_file = bem_dir / 'sample-5120-5120-5120-bem-sol.fif'
        trans_file = meg_dir / 'sample_audvis_raw-trans.fif'

        if fwd_file.exists():
            fwd = mne.read_forward_solution(fwd_file)
        else:
            src_ = _mne_source_space(subject, src_tag, subjects_dir)
            fwd = mne.make_forward_solution(epochs.info, trans_file, src_, bem_file)
            mne.write_forward_solution(fwd_file, fwd)

        cov_file = meg_dir / 'sample_audvis-cov.fif'
        cov = mne.read_cov(cov_file)
        inv = mn.make_inverse_operator(epochs.info, fwd, cov, loose=loose, depth=None, fixed=fixed)
        mne.minimum_norm.write_inverse_operator(inv_file, inv)
    ds.info['inv'] = inv

    stcs = mn.apply_inverse_epochs(epochs, inv, 1. / (snr ** 2), method, pick_ori=pick_ori)
    ds['src'] = load.mne.stc_ndvar(stcs, subject, src_tag, subjects_dir, method, fixed)
    if stc:
        ds['stc'] = stcs

    if fsaverage:
        src_to = _mne_source_space('fsaverage', src_tag, subjects_dir)
        morph = mne.compute_source_morph(stcs[0], subject, 'fsaverage', subjects_dir, src_to=src_to)
        stcs_m = [morph.apply(stc) for stc in stcs]
        ds['srcm'] = load.mne.stc_ndvar(stcs_m, 'fsaverage', src_tag, subjects_dir, method, fixed)

    return ds


def get_ndvar(case=0, time=100, frequency=8, cat=0, sensor=0, name='ndvar'):
    dims = []
    if case:
        dims.append(Case(case))
    if time:
        dims.append(UTS(-0.1, 0.01, time))
    if frequency:
        dims.append(Scalar('frequency', np.logspace(2, 3.5, frequency)))
    if cat:
        dims.append(Categorial('cat', string.ascii_lowercase[:cat]))
    if sensor:
        dims.append(get_sensor(sensor))
    shape = [len(dim) for dim in dims]
    x = np.random.normal(0, 1, shape)
    return NDVar(x, dims, name)


def get_uts(utsnd=False, seed=0, nrm=False, vector3d=False):
    """Create a sample Dataset with 60 cases and random data.

    Parameters
    ----------
    utsnd : bool
        Add a sensor by time NDVar (called 'utsnd').
    seed : None | int
        If not None, call ``numpy.random.seed(seed)`` to ensure replicability.
    nrm : bool
        Add a nested random effect Factor "nrm" (nested in ``A``).
    vector3d : bool
        Add a space x time vector time series as ``v3d``.

    Returns
    -------
    ds : Dataset
        Datasets with data from random distributions.
    """
    random = np.random if seed is None else np.random.RandomState(seed)

    ds = Dataset()

    # add a model
    ds['A'] = Factor(['a0', 'a1'], repeat=30)
    ds['B'] = Factor(['b0', 'b1'], repeat=15, tile=2)
    ds['rm'] = Factor(('R%.2i' % i for i in range(15)), tile=4, random=True)
    ds['ind'] = Factor(('R%.2i' % i for i in range(60)), random=True)

    # add dependent variables
    rm_var = np.tile(random.normal(size=15), 4)
    y = np.hstack((random.normal(size=45), random.normal(1, size=15)))
    y += rm_var
    ds['Y'] = Var(y)
    ybin = random.randint(0, 2, size=60)
    ds['YBin'] = Factor(ybin, labels={0: 'c1', 1: 'c2'})
    ycat = random.randint(0, 3, size=60)
    ds['YCat'] = Factor(ycat, labels={0: 'c1', 1: 'c2', 2: 'c3'})

    # add a uts NDVar
    time = UTS(-.2, .01, 100)
    y = random.normal(0, .5, (60, len(time)))
    y += rm_var[:, None]
    y[:15, 20:60] += np.hanning(40) * 1  # interaction
    y[:30, 50:80] += np.hanning(30) * 1  # main effect
    ds['uts'] = NDVar(y, ('case', time))

    # add sensor NDVar
    if utsnd:
        y = random.normal(0, 1, (60, 5, len(time)))
        y += rm_var[:, None, None]
        # add interaction
        win = np.hanning(50)
        y[:15, 0, 50:] += win * 3
        y[:15, 1, 50:] += win * 2
        y[:15, 4, 50:] += win
        # add main effect
        y[30:, 2, 25:75] += win * 2.5
        y[30:, 3, 25:75] += win * 1.5
        y[30:, 4, 25:75] += win
        # add spectral effect
        freq = 15.0  # >= 2
        x = np.sin(time.times * freq * 2 * np.pi)
        for i in range(30):
            shift = random.randint(0, 100 / freq)
            y[i, 2, 25:75] += 1.1 * win * x[shift: 50 + shift]
            y[i, 3, 25:75] += 1.5 * win * x[shift: 50 + shift]
            y[i, 4, 25:75] += 0.5 * win * x[shift: 50 + shift]

        sensor = get_sensor(5)
        dims = ('case', sensor, time)
        ds['utsnd'] = NDVar(y, dims, info=_info.for_eeg())

    # nested random effect
    if nrm:
        ds['nrm'] = Factor([f'{a}{i:02}' for a in 'AB' for _ in range(2) for
                            i in range(15)], random=True)

    if vector3d:
        x = random.normal(0, 1, (60, 3, 100))
        # main effect
        x[:30, 0, 50:80] += np.hanning(30) * 0.7
        x[:30, 1, 50:80] += np.hanning(30) * -0.5
        x[:30, 2, 50:80] += np.hanning(30) * 0.3
        ds['v3d'] = NDVar(x, (Case, Space('RAS'), time))

    return ds


def get_uv(seed=0, nrm=False, vector=False):
    """Dataset with random univariate data

    Parameters
    ----------
    seed : None | int
        Seed the numpy random state before generating random data.
    nrm : bool
        Add a nested random-effects variable (default False).
    vector : bool
        Add a 3d vector variable as ``ds['v']`` (default ``False``).
    """
    random = np.random if seed is None else np.random.RandomState(seed)

    ds = permute([('A', ('a1', 'a2')),
                  ('B', ('b1', 'b2')),
                  ('rm', ['s%03i' % i for i in range(20)])])
    ds['rm'].random = True
    ds['intvar'] = Var(random.randint(5, 15, 80))
    ds['intvar'][:20] += 3
    ds['fltvar'] = Var(random.normal(0, 1, 80))
    ds['fltvar'][:40] += 1.
    ds['fltvar2'] = Var(random.normal(0, 1, 80))
    ds['fltvar2'][40:] += ds['fltvar'][40:].x
    ds['index'] = Var(np.repeat([True, False], 40))
    if nrm:
        ds['nrm'] = Factor(['s%03i' % i for i in range(40)], tile=2, random=True)
    if vector:
        x = random.normal(0, 1, (80, 3))
        x[:40] += [.3, .3, .3]
        ds['v'] = NDVar(x, (Case, Space('RAS')))
    return ds


def get_sensor(n):
    assert n == 5
    locs = np.array([[-1.0, 0.0, 0.0],
                     [0.0, 1.0, 0.0],
                     [1.0, 0.0, 0.0],
                     [0.0, -1.0, 0.0],
                     [0.0, 0.0, 1.0]])
    sensor = Sensor(locs, sysname='test_sens')
    sensor.set_adjacency(connect_dist=1.75)
    return sensor


def setup_samples_experiment(
        dst: PathArg,
        n_subjects: int = 3,
        n_tasks: int = 1,
        n_segments: int = 4,
        name: str = 'SampleExperiment',
        pick: str = 'mag',
):
    """Setup up file structure for the ``SampleExperiment`` class

    Corresponding to the pipeline in ``examples/experiment``.

    Parameters
    ----------
    dst
        Path. ``dst`` should exist, a new folder called ``name`` will be
        created within ``dst``.
    n_subjects
        Number of subjects.
    n_segments
        Number of data segments to include in each file.
    n_sessions
        Number of sessions.
    n_visits
        Number of visits.
    name
        Name for the directory for the new experiment (default
        ``'SampleExperiment'``).
    mris
        Set up MRIs.
    mris_only
        Only create MRIs, skip MEG data (add MRIs to existing experiment data).
    pick
        Pick a certain channel type (``''`` to copy all channels).
    """
    # input paths
    data_path = mne.datasets.sample.data_path()
    fsaverage_path = mne.datasets.fetch_fsaverage()
    raw_fname = join(data_path, "MEG", "sample", "sample_audvis_raw.fif")
    event_id = {
        "Auditory/Left": 1,
        "Auditory/Right": 2,
        "Visual/Left": 3,
        "Visual/Right": 4,
        "Smiley": 5,
        "Button": 32,
    }

    # read raw
    raw = mne.io.read_raw_fif(raw_fname)
    raw.info['line_freq'] = 60
    raw.info['bads'] = []
    sfreq = raw.info['sfreq']

    # segmentation
    n_recordings = n_subjects * n_tasks
    events = mne.find_events(raw)
    events[:, 0] -= raw.first_samp
    segs = []
    n = 0
    t_start = 0
    for sample, _, trigger in events:
        if trigger == 5:  # smiley
            n += 1
        if n == n_segments:
            t = sample / sfreq
            segs.append((t_start, t))
            if len(segs) == n_recordings:
                break
            t_start = t
            n = 0
    else:
        raise ValueError("Not enough data in sample raw. Try smaller ns.")

    # generate subject and task names
    subjects = [f'R{s_id:04}' for s_id in range(n_subjects)]
    if n_tasks > 1:
        tasks = ['sample%i' % (i + 1) for i in range(n_tasks)]
    else:
        tasks = ['sample']

    # output paths
    dst = Path(dst).expanduser().resolve()
    root = dst / name
    root.mkdir()
    bids_path = BIDSPath(root=root)

    for subject in subjects:
        for task in tasks:
            start, stop = segs.pop()
            raw_ = raw.copy().crop(start, stop)
            raw_.load_data()
            if pick:
                raw_.pick_types(pick, stim=True, exclude=[])
            bids_path.update(subject=subject, task=task, datatype='meg')
            write_raw_bids(
                raw=raw_,
                bids_path=bids_path,
                event_id=event_id,
                overwrite=True,
                allow_preload=True,
                format='FIF',
            )

    # freesurfer
    mri_sdir = root / 'derivatives' / 'freesurfer'
    mri_sdir.mkdir(parents=True)
    # copy rudimentary fsaverage
    surf_names = ['inflated', 'white', 'orig', 'orig_avg', 'curv', 'sphere']
    files = {
        'bem': ['fsaverage-head.fif', 'fsaverage-inner_skull-bem.fif'],
        'label': ['lh.aparc.annot', 'rh.aparc.annot'],
        'surf': [f'{hemi}.{name}' for hemi, name in product(['lh', 'rh'], surf_names)],
        'mri': [],
    }
    dst_s_dir = mri_sdir / 'fsaverage'
    dst_s_dir.mkdir()
    # from fsaverage
    for dir_name, file_names in files.items():
        src_dir = fsaverage_path / dir_name
        dst_dir = dst_s_dir / dir_name
        dst_dir.mkdir()
        for file_name in file_names:
            shutil.copy(src_dir / file_name, dst_dir / file_name)
    # source space
    src_src = fsaverage_path / 'bem' / 'fsaverage-ico-3-src.fif'
    src_dst = dst_s_dir / 'bem' / 'fsaverage-ico-3-src.fif'
    if not src_src.exists():
        src = mne.setup_source_space('fsaverage', 'ico1', subjects_dir=fsaverage_path.parent)
        src.save(src_src)
    shutil.copy(src_src, src_dst)

    # trans
    trans = mne.Transform(4, 5, [[ 0.9998371,  -0.00766024,  0.01634169,  0.00289569],
                                 [ 0.00933457,  0.99443108, -0.10497498, -0.0205526 ],
                                 [-0.01544655,  0.10511042,  0.9943406,  -0.04443745],
                                 [ 0.,          0.,          0.,          1.        ]])
    for subject in subjects:
        mne.scale_mri('fsaverage', subject, 1., subjects_dir=mri_sdir, skip_fiducials=True, labels=False)
        trans_dir = root / 'derivatives' / 'trans' / subject
        trans_dir.mkdir(parents=True)
        trans.save(str(trans_dir / f'{subject}_trans.fif'))
